from cryptography.fernet import Fernet


def _chunks(data, size):
    for i in range(0, len(data), size):
        yield data[i:i + size]


def format_encrypted_data(data):
    res = ''
    for i in _chunks(data.decode(), 64):
        res += i + '\n'
    return res


def generate_symetric_key(fname=None):
    key = Fernet.generate_key()
    if fname:
        with open(fname, 'wb') as f:
            f.write(key)
    return Fernet(key)


def load_symetric_key(fname):
    with open(fname, 'rb') as f:
        key = f.read()
        return Fernet(key)


def get_or_create_symetric_key(fname):
    try:
        return load_symetric_key(fname)
    except FileNotFoundError:
        return generate_symetric_key(fname)


# def generate_asymetric_keys(public_fname=None, private_fname=None):
#     public_key, private_key = rsa.newkeys(2048)
#
#     if public_fname:
#         with open(public_fname, 'wb') as f:
#             f.write(public_key.save_pkcs1('PEM'))
#
#     if private_fname:
#         with open(private_fname, 'wb') as f:
#             f.write(private_key.save_pkcs1('PEM'))
#
#     return public_key, private_key
#
#
# def load_public_key(fname):
#     with open(fname, 'rb') as f:
#         public_key = rsa.PublicKey.load_pkcs1(f.read())
#         return public_key
#
#
# def load_private_key(fname):
#     with open(fname, 'rb') as f:
#         private_key = rsa.PrivateKey.load_pkcs1(f.read())
#         return private_key



from contextlib import contextmanager
import os
import textwrap
import click
import json


def local_appdata_dir():
    """Get the application data directory.
    """
    return click.get_app_dir('seeqret', roaming=False)


def roaming_appdata_dir():
    """Get the application data directory.
    """
    return click.get_app_dir('seeqret', roaming=True)


def remove_file_if_exists(fname):
    """Remove a file if it exists.
    """
    if os.path.exists(fname):
        os.remove(fname)


def move_file_and_overwrite(src, dst):
    """Move a file. Remove the destination if it exists.
    """
    if os.path.exists(dst):
        os.remove(dst)
    os.rename(src, dst)


def write_file(fname, content):
    """Write content to a file.
    """
    with open(fname, 'w') as f:
        f.write(textwrap.dedent(content))


def read_file(fname):
    """Read content from a file.
    """
    with open(fname, 'r') as f:
        return f.read()


def read_json(fname):
    """Read json content from a file.
    """
    with open(fname, 'r') as f:
        return json.load(f)


def read_binary_file(fname):
    """Read content from a file.
    """
    with open(fname, 'rb') as f:
        return f.read()


def is_writable(dirname):
    """Check if a directory is writable.
    """
    return os.access(dirname, os.W_OK)


def is_readable(dirname):
    """Check if a directory is readable.
    """
    return os.access(dirname, os.R_OK)


def remove_directory(dirname):
    """Remove a directory, recursively.
    """
    if os.path.exists(dirname):
        for root, dirs, files in os.walk(dirname, topdown=False):
            for name in files:
                os.remove(os.path.join(root, name))
            for name in dirs:
                os.rmdir(os.path.join(root, name))
        os.rmdir(dirname)


def write_binary_file(fname, content):
    """Write content to a file.
    """
    with open(fname, 'wb') as f:
        f.write(content)


@contextmanager
def cd(path):
    """Change the current working directory.
    """
    old_dir = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_dir)


def run(cmd, echo=True):
    if echo:
        click.secho(f"    > {cmd}", fg='blue')
    res = os.popen(cmd).read()
    if echo:
        for line in res.split('\n'):
            click.secho(f"      {line}", fg='green')
    return res


def attrib_cmd(vault_dir, cmd=""):
    return run(f"attrib {cmd} {vault_dir}").strip()


def is_encrypted(dirname):
    """Check if a directory is encrypted.
    """
    if os.name == 'nt':
        return run(f"cipher /c {dirname}").find(f'E {dirname}') != -1
    return False



import os
from pathlib import Path

import click

# from .context import Context
from . import seeqret_init, seeqret_add
from .utils import is_writable, cd, read_json

DIRNAME = Path(__file__).parent

# os.environ['SEEQRET'] = os.getcwd() + r'\seeqret'


@click.group()
# @click.pass_context
def cli():
    pass


@cli.command()
def upgrade():
    """Upgrade the database to the latest version
    """
    with cd(os.environ['SEEQRET']):
        seeqret_init.upgrade_db()


@cli.command()
def list():
    """List the contents of the vault
    """
    with cd(os.environ['SEEQRET']):
        return seeqret_add.list_secrets()


@cli.command()
def users():
    """List the users in the vault
    """
    # print("SEEQRET:DIR:", os.environ['SEEQRET'])
    with cd(os.environ['SEEQRET']):
        seeqret_add.list_users()


@cli.command()
@click.argument('to')
def export(to):
    """Export the vault to a user
    """
    with cd(os.environ['SEEQRET']):
        seeqret_add.export_secrets(to)


@cli.command()
@click.argument('fname')
def import_file(fname):
    """Import a vault from a file
    """
    text = read_json(fname)
    with cd(os.environ['SEEQRET']):
        seeqret_add.import_secrets(text)


@cli.command()
@click.argument(
    'dir',
    default='.',
    # prompt='Directory to initialize seeqret in',
)
@click.option(
    '--user',
    prompt=True,
    default=lambda: f"{os.environ['USERDOMAIN']}\\{os.environ['USERNAME']}"
)
@click.option('--email', prompt=True)
@click.option('--pubkey', default=None)
@click.option('--key', default=None)
def init(dir, user, email, pubkey=None, key=None):
    """Initialize a new vault
    """
    dirname = Path(dir).resolve()
    vault_dir = dirname / 'seeqret'

    # we want to create dirname / seeqret

    if not dirname.exists():
        click.echo(f'The parent of the vault: {dirname} must exist.')
        return

    if not is_writable(dirname):
        click.echo(f'The parent of the vault: {dirname} must be writable.')
        return

    if vault_dir.exists():
        if not is_writable(vault_dir):
            click.echo(
                f'The vault: {vault_dir} exists and is not writeable, '
                'you must delete it manually.'
            )
            return
        click.confirm(
            f'The vault: {vault_dir} already exists, overwrite contents?',
            default=True, abort=True)
        # remove_directory(vault_dir)

    seeqret_init.secrets_init(dirname, user, email, pubkey, key)


@cli.group()
def add():
    """Add a new secret, key or user
    """
    pass


@add.command()
@click.option('--url', prompt=True)
@click.option('--username', prompt=True)
@click.option('--email', prompt=True)
def user(url, username, email):
    """Add a new user to the vault from a public key.

       If the public key is on github, the url is the raw url, e.g.

       https://raw.githubusercontent.com/username/project/refs/heads/main/public.key
    """
    click.echo(f'Adding a new user, from {url}')
    with cd(os.environ['SEEQRET']):
        click.secho(f'Fetching public key: {url}', fg='blue')
        pubkey = seeqret_add.fetch_pubkey_from_url(url)
        seeqret_add.add_user(pubkey, username, email)


@add.command()
@click.argument('name')
@click.argument('value')
@click.option('--app', default='*')
@click.option('--env', default='*')
def key(name: str, value: str, app: str = None, env: str = None):
    """Add a new key/value pair.
    """
    click.echo(
        f'Adding a new key: {name}, value: {value}, app: {app}, env: {env}'
    )

    with cd(os.environ['SEEQRET']):
        seeqret_add.add_key(name, value, app, env)



# import os
#
# from click.testing import CliRunner
# from pathlib import Path
#
# from seeqret import seeqret_init, cd
# from seeqret.main import (
#     init,
#     list,
#     users,
#     user,
#     key,
# )
# from seeqret.utils import run
#
#
# def test_cli(monkeypatch):
#     runner = CliRunner()
#
#     with runner.isolated_filesystem():
#         tdir = Path(os.getcwd())
#         monkeypatch.setenv('SEEQRET', str(tdir / 'seeqret'))
#         init(str(tdir), 'test', 'test@example.com')
#
#         with cd(tdir):
#             print("TDIR:", tdir)
#             print("CURDIR:", os.getcwd())
#             assert tdir == Path(os.getcwd())
#
#             # result = runner.invoke(init, [
#             #     tdir,
#             #     # '--url', 'https://raw.githubusercontent.com/tkbeorg/tkbe/refs/heads/main/public.key',
#             #     '--user', 'test',
#             #     '--email' 'test@example.com',
#             # ])
#             # print("INIT:OUTPUT:", result.output)
#             run('tree .')
#             # assert listdir contains seeqret
#
#             # print("LIST:", os.path.listdir(tdir / 'seeqret')
#
#             result = runner.invoke(list)
#             assert result.output.strip() == ''
#
#             result = runner.invoke(users)
#             assert len(result.output) == 42
#             print("CLIRUNNER:OUTPUT:", result.output)
#             # print("CLIRUNNER:EXIT:CODE:", result.exit_code)
#             assert 'test@example.com (admin)' in result.output
#
#             result = runner.invoke(user,[
#                 '--username', 'tkbe',
#                 '--email', 'bjorn@tkbe.org',
#                 '--url', 'https://raw.githubusercontent.com/tkbeorg/tkbe/refs/heads/main/public.key'
#             ])
#             assert 'bjorn@tkbe.org' in result.output
#
#             result = runner.invoke(users)
#             assert len(result.output) == 42
#             assert 'bjorn@tkbe.org' in result.output
#
#             result = runner.invoke(key, [
#                 'FOO', 'bar'
#             ])
#             assert 'FOO' in result.output



from nacl.hash import sha256
from nacl.public import PrivateKey, Box, PublicKey
from nacl import encoding
from nacl.signing import SigningKey

from seeqret.utils import write_binary_file, read_binary_file


def generate_private_key(fname=None) -> PrivateKey:
    pkey = PrivateKey.generate()
    if fname:
        write_binary_file(
            fname,
            encoding.Base64Encoder.encode(pkey._private_key),
        )
    return pkey


def private_key(string: bytes) -> PrivateKey:
    """Bytes to private key.
    """
    return PrivateKey(string, encoder=encoding.Base64Encoder)


def public_key(string: str | bytes) -> PublicKey:
    """Bytes to public key.
    """
    if isinstance(string, str):
        string = string.encode('ascii')
    return PublicKey(string, encoder=encoding.Base64Encoder)


def load_private_key(fname: str) -> PrivateKey:
    """Load private key from file.
    """
    return private_key(read_binary_file(fname))


def load_public_key(fname: str) -> PublicKey:
    """Load public key from file.
    """
    return public_key(read_binary_file(fname))


def save_public_key(fname: str, pkey: PrivateKey) -> bytes:
    pubkey = encoding.Base64Encoder.encode(bytes(pkey.public_key))
    write_binary_file(fname, pubkey)
    return pubkey


def save_private_key(fname: str, pkey: PrivateKey) -> bytes:
    pkey = encoding.Base64Encoder.encode(bytes(pkey._private_key))
    write_binary_file(fname, pkey)
    return pkey


def asymetric_encrypt_string(string: str,
                             sender_private_key: PrivateKey,
                             recipient_public_key: PublicKey) -> str:
    box = Box(
        private_key=sender_private_key,
        public_key=recipient_public_key
    )
    val = string.encode('utf-8')
    return encoding.Base64Encoder.encode(box.encrypt(val)).decode('ascii')


def asymetric_decrypt_string(string: str,
                             receiver_private_key: PrivateKey,
                             sender_public_key: PublicKey) -> str:
    box = Box(
        private_key=receiver_private_key,
        public_key=sender_public_key
    )
    val = string.encode('ascii')
    return box.decrypt(encoding.Base64Encoder.decode(val)).decode('ascii')


def sign_message(string: bytes):
    signing_key = SigningKey.generate()
    signed = signing_key.sign(string)
    verify_key = signing_key.verify_key
    verify_key_bytes = verify_key.encode()
    return verify_key_bytes, signed


def hash_message(string: bytes) -> str:
    return sha256(string, encoder=encoding.Base64Encoder).decode('ascii')
